module NPhotonsMakieExt

using NPhotons, Makie, LinearAlgebra, Interpolations, StatsBase
# const nph = NPhotons
import Makie.SpecApi as S

Makie.used_attributes(::Type{<:Plot}, x::NPhotons.AtomVolume) = (:ngrid, :r, :itp, :normalization, :scale, :cutoff)

function Makie.convert_arguments(P::Type{<:Volume}, xs, ys, zs, f::NPhotons.AtomVolume; itp = 0, normalization = :maximum, scale = identity, cutoff = Inf)
    fs = [f(x, y, z, :real, cutoff = cutoff) for x in xs, y in ys, z in zs]
    if normalization == :maximum
        fs ./= maximum(fs)
    elseif normalization == :mean
        fs ./= sum(fs) / length(fs)
    elseif normalization == :gaussian
        fs ./= minimum(f.heights ./ f.widths.^3) / (2pi)^(3/2)
    end
    fs .= scale.(fs)

    if itp > 0
        fitp = cubic_spline_interpolation((xs, ys, zs), fs)
        xs = range(extrema(xs)..., length = length(xs) * itp)
        ys = range(extrema(ys)..., length = length(ys) * itp)
        zs = range(extrema(zs)..., length = length(zs) * itp)
        fs = [fitp(x, y, z) for x in xs, y in xs, z in xs]
    end

    (Makie.to_interval.((xs, ys, zs))..., Float32.(fs))
end

function Makie.convert_arguments(P::Type{<:Volume}, f::NPhotons.AtomVolume; ngrid = 50, r = nothing, itp = 0, normalization = :maximum, scale = identity, cutoff = Inf)
    r = isnothing(r) ? (maximum(x -> norm(x, Inf), f.positions) + 3maximum(f.widths)) : r
    xs = range(-r, r, length = ngrid)
    fs = [f(x, y, z, :real, cutoff = cutoff) for x in xs, y in xs, z in xs]

    if normalization == :maximum
        fs ./= maximum(fs)
    elseif normalization == :mean
        fs ./= sum(fs) / length(fs)
    elseif normalization == :gaussian
        fs ./= minimum(f.heights ./ f.widths.^3) / (2pi)^(3/2)
    end
    
    if itp > 0
        fitp = cubic_spline_interpolation((xs, xs, xs), fs)
        xs = range(-r, r, length = itp * ngrid)
        fs = [fitp(x, y, z) for x in xs, y in xs, z in xs]
    end
    fs .= scale.(fs)

    (Makie.to_interval.((xs, xs, xs))..., Float32.(fs))
end

# function Makie.convert_arguments(P::Type{<:Volume}, f::NPhotons.AtomVolume; kwargs...)
#     Makie.convert_arguments(Volume, f; kwargs...)
# end

# function Makie.plot!(plot::Volume{<:Tuple{<:NPhotons.AtomVolume}})
#     println(plot)
#     valid_attributes = Makie.shared_attributes(plot, Volume)
#     volume!(plot, valid_attributes, plot[1])
# end

function Makie.convert_arguments(P::Type{<:Scatter}, f::NPhotons.AtomVolume)
    Makie.convert_arguments(P, f.positions)
end

cam3dfixed!(scene; kwargs...) = cam3d!(scene; zoom_shift_lookat = false, kwargs...)

function liftevery(innode, dt)
    t = Observable(time_ns())
    outnode = Observable(innode[])
    on(innode) do val
        if time_ns() - t[] > dt * 1e9
            t[] = time_ns()
            outnode[] = val
        end
    end
    outnode
end

function linkCams!(scenes)
    for scene in scenes
        scene.camera = copy(scenes[1].camera)
        on(scene.camera_controls.eyeposition) do ep
            for scene2 in scenes
                scene2.camera_controls.eyeposition.val = ep
            end
        end
        on(scene.camera_controls.lookat) do ep
            for scene2 in scenes
                scene2.camera_controls.lookat.val = ep
            end
        end
        on(scene.camera_controls.upvector) do ep
            for scene2 in scenes
                scene2.camera_controls.upvector.val = ep
            end
        end
        on(scene.camera_controls.fov) do ep
            for scene2 in scenes
                scene2.camera_controls.fov.val = ep
            end
        end
    end
end

function NPhotons.monitor(state::NPhotons.Annealer, object = nothing; aligndt = 5, volumedt = 1, dt = 0, showinfo = false, showwidths = false, yscale = identity, centersize = 0, showres = false, partialylimits = false, showobject = true, includezero = true, kwargs...)
    if isnothing(object)
        showobject = false
        object = NPhotons.AtomVolume(1, 10, 1)
    end
    fig = Figure(; kwargs...)
    state = liftevery(state, dt)
    
    fs = lift(NPhotons.getstructures, liftevery(state, volumedt))
    Rs = lift(f -> NPhotons.align.(f, object, ntries = 10000), liftevery(fs, aligndt))

    volgrid = GridLayout(fig[1,1])
    rowsize!(fig.layout, 1, Relative(0.5))
    os = NPhotons.getstructures(object)

    isorange = Observable(2.0)

    if showobject
        for (i, o) in enumerate(os)
            ls = LScene(volgrid[1,i]; show_axis = false);
            cam3dfixed!(ls, clipping_mode = :bbox_relative, far = 2)
            volume!(ls, o, colormap = :ice, normalization = :gaussian, scale = log10, isorange = isorange)
            if centersize != 0
                meshscatter!(ls, o.positions, color = :black, markersize = centersize .* o.widths .* o.heights ./ maximum(o.heights), overdraw = true)
            end
        end
    end

    for i in 1:length(fs[])
        ls = LScene(volgrid[1 + (length(os) > 1), i + showobject * (length(os) <= 1)]; show_axis = false);
        cam3dfixed!(ls, clipping_mode = :bbox_relative, far = 2)
        f = lift((R, f) -> R[i] * f[i], Rs, fs)
        volume!(ls, f, colormap = :viridis, normalization = :gaussian, scale = log10, isorange = isorange)
        if centersize != 0
            meshscatter!(ls, lift(x -> x.positions, f), color = :black, markersize = lift(x -> centersize .* x.widths .* x.heights ./ maximum(x.heights), f), overdraw = true)
        end
    end   

    linkCams!([ls.scene for ls in contents(volgrid[:,:])])

    sgrid = GridLayout(volgrid[2 + (length(os) > 1), :])
    isoslider = Slider(sgrid[1,1], range = LinRange(1, -5, 10000))
    Label(sgrid[1,2], @lift string(round(exp10($(isoslider.value)), sigdigits = 2)))
    on(isoslider.value) do v isorange[] = 1 - v end
    set_close_to!(isoslider, -1)

    # ls2 = LScene(fig[1,1][1,2]; show_axis = false);
    # cam3dfixed!(ls2, clipping_mode = :bbox_relative, far = 2)
    # volume!(ls2, lift((R, s) -> R * s.configuration, Rs, liftevery(state, volumedt)))
    
    

    egrid = GridLayout(fig[2,1])
    ax = if yscale == :pseudolog
        yticks = [-10 .^ (0:10); 0; 10 .^ (0:10)]
        yticklabels = [[rich("-10", superscript(string(i))) for i in 0:10]; rich("0"); 
                       [rich("10", superscript(string(i))) for i in 0:10]]
        Axis(egrid[1,1], 
            yscale = ReversibleScale(x -> asinh(2*sqrt(6)*x), x -> sinh(x)/(2*sqrt(6))),
            yticks = (yticks, yticklabels), subtitle = "-log(p)"
        )
    else
        Axis(egrid[1, 1], subtitle = "-log(p)")
    end
    includezero && hlines!(ax, [0, 1], color = :transparent)
    energies = lift(state) do s
        if yscale == :log
            offset = max(1, maximum(s.energies[end-end÷10:end]) - minimum(s.energies[end-end÷10:end]))
            log10.(s.energies .- minimum(s.energies) .+ offset)
        else
            s.energies
        end
    end
    lines!(ax, energies)
    on(energies, priority = -1) do s
        if partialylimits && length(s) > 10
            xlims!(ax, nothing, nothing)
            ymin, ymax = extrema(last.(s)[end÷10:end])
            ymin = min(0, ymin); ymax = max(0, ymax); 
            ylims!(ax, ymin - 0.1 * (ymax - ymin), ymax + 0.1 * (ymax - ymin))
        else
            autolimits!(ax)
        end
    end

    c = state[].configuration
    if (c isa Array || c isa Tuple) && length(c) == 2 && c[2] isa Vector{<:AbstractFloat}
        ax1 = Axis(egrid[1,2], subtitle = "weights")
        for i in 1:length(c[2])
            weights = lift(s -> [h[2][i] / sum(h[2]) for h in s.history], state)
            lines!(ax1, weights)
        end
        hlines!(ax1, 0, color = :transparent)
        on(state) do s
            autolimits!(ax1)
        end
        # axislegend(ax1, framevisible = false, position = :lb)
    elseif showres
        axres = Axis(egrid[1,2], subtitle = "resolution", yticks = WilkinsonTicks(5))
        resolutions = Observable(Tuple{Int, Float64}[])
        lines!(axres, resolutions)
        # scatter!(axres, [1,1], [0, 1], color = :transparent)
        on(state) do s
            push!(resolutions[], (s.nsteps, NPhotons.fscResolution(Rs[][1] * fs[][1], os[1], precision = 17)))
            notify(resolutions)
            autolimits!(axres)
        end
    end

    if showinfo
        infogrid = GridLayout(fig[3,1])
        ax2 = Axis(infogrid[1,1], 
            yscale = log10, yticks = LogTicks(-10:1:10), 
            yminorticksvisible=true, yminorticks = IntervalsBetween(9), subtitle = "step size"
        );
        for i in 1:length(state[].proposers)
            stepsizes = Observable(Point2f[])
            lines!(ax2, stepsizes)
            scatter!(ax2, [1,1], [1, 10], color = :transparent)
            on(state) do s
                push!(stepsizes[], Point2f(s.nsteps, s.proposers[i].stepsize))
                notify(stepsizes)
                autolimits!(ax2)
            end
        end

        ax3 = Axis(infogrid[1,2], subtitle = "acceptance rate")
        for i in 1:length(state[].proposers)
            acceptancerates = Observable(Tuple{Int, Float64}[])
            lines!(ax3, acceptancerates)
            scatter!(ax3, [1,1], [0, 1], color = :transparent)
            on(state) do s
                push!(acceptancerates[], (s.nsteps, NPhotons.acceptanceRatio(s, n = 100)[i]))
                notify(acceptancerates)
                autolimits!(ax3)
            end
        end

        if showwidths
            ax4 = Axis(infogrid[1,3], subtitle = "widths")
            widths = lift(s -> [NPhotons.getstructures(h)[1].widths[1] for h in s.history], state)
            lines!(ax4, widths)
            on(state) do s
                autolimits!(ax4)
            end
        end
        Label(infogrid[2,:], lift(s -> "T = $(round(s.temperature, sigdigits = 5))   E = $(round(s.energies[end], sigdigits = 5))   d = $(round.(getfield.(s.proposers, :stepsize), sigdigits = 5))   nsteps = $(s.nsteps)", state))
    end
    state[] = state[]

    fig
end

function NPhotons.monitor(chain::Observable{<:AbstractArray{<:NPhotons.AnnealingState}}, object = nothing; dt = 0, aligndt = 5, volumedt = 1, showobject = true, centersize = 0, yscale = :identity, includezero = true, partialylimits = false, showinfo = true, cutoff = Inf, size = (1200, 1200), kwargs...)
    fig = Figure(; size = size, figure_padding = 25, kwargs...)
    dtchain = liftevery(chain, dt)

    if isnothing(object)
        showobject = false
        object = NPhotons.AtomVolume(1, 10, 1)
    end

    fs = lift(c -> NPhotons.getstructures(c[end].conf), liftevery(dtchain, volumedt))
    Rs = lift(f -> NPhotons.align.(f, object, ntries = 2000), liftevery(fs, aligndt))
    os = NPhotons.getstructures(object)
    

    volgrid = GridLayout(fig[1,1])
    rowsize!(fig.layout, 1, Relative(0.5))
    isorange = Observable(2.0)

    if showobject
        for (i, o) in enumerate(os)
            ls = LScene(volgrid[1,i]; show_axis = false);
            cam3dfixed!(ls, clipping_mode = :bbox_relative, far = 2)
            volume!(ls, o, colormap = :ice, normalization = :gaussian, scale = log10, isorange = isorange, cutoff = cutoff)
            if centersize != 0
                meshscatter!(ls, o.positions, color = :black, markersize = centersize .* o.heights ./ maximum(o.heights), overdraw = true)
            end
        end
    end

    for i in 1:length(fs[])
        ls = LScene(volgrid[1 + (length(os) > 1), i + showobject * (length(os) <= 1)]; show_axis = false);
        cam3dfixed!(ls, clipping_mode = :bbox_relative, far = 2)
        f = lift((R, f) -> R[i] * f[i], Rs, fs)
        volume!(ls, f, colormap = :viridis, normalization = :gaussian, scale = log10, isorange = isorange, cutoff = cutoff)
        if centersize != 0
            meshscatter!(ls, lift(x -> x.positions, f), color = :black, markersize = lift(x -> centersize .* x.heights ./ maximum(x.heights), f), overdraw = true)
        end
    end   

    linkCams!([ls.scene for ls in contents(volgrid[:,:])])

    isoslider = Slider(volgrid[:, end+1], horizontal = false, range = LinRange(1, -5, 10000))
    on(isoslider.value) do v isorange[] = 1 - v end
    set_close_to!(isoslider, -1)

    egrid = GridLayout(fig[2,1])
    ax = if yscale == :pseudolog
        yticks = [-10 .^ (0:10); 0; 10 .^ (0:10)]
        yticklabels = [[rich("-10", superscript(string(i))) for i in 0:10]; rich("0"); 
                       [rich("10", superscript(string(i))) for i in 0:10]]
        Axis(egrid[1,1], 
            yscale = ReversibleScale(x -> asinh(2*sqrt(6)*x), x -> sinh(x)/(2*sqrt(6))),
            yticks = (yticks, yticklabels), subtitle = "-log(p)"
        )
    else
        Axis(egrid[1, 1], subtitle = "-log(p)")
    end
    includezero && hlines!(ax, [0, 1], color = :transparent)
    energies = lift(dtchain) do c
        es = (s -> s.energy).(c)
        if yscale == :log
            tail = es[max(1, end-100):end]
            offset = max(1, std(tail, eweights(length(tail), 0.1)))
            Point2.(getnsteps.(c) .- c[1].nsteps, log10.(es .- minimum(es) .+ offset))
        else
            Point2.(getnsteps.(c) .- c[1].nsteps, es)
        end
    end
    lines!(ax, energies)
    on(energies, priority = -1) do s
        if partialylimits && length(s) > 10
            xlims!(ax, nothing, nothing)
            ymin, ymax = extrema(last.(s)[end÷10:end])
            ymin = min(0, ymin); ymax = max(0, ymax); 
            ylims!(ax, ymin - 0.1 * (ymax - ymin), ymax + 0.1 * (ymax - ymin))
        else
            autolimits!(ax)
        end
    end

    if showinfo
        infogrid = GridLayout(fig[3,1])
        ax2 = Axis(infogrid[1,1], 
            yscale = log10, yticks = LogTicks(-10:1:10), 
            yminorticksvisible=true, yminorticks = IntervalsBetween(9), subtitle = "step size"
        );
        scatter!(ax2, [1,1], [1, 10], color = :transparent)
        if dtchain[][end].proposalParams isa Real
            lines!(ax2, lift(c -> [s.proposalParams for s in c], dtchain))
        else
            for i in 1:length(dtchain[][end].proposalParams.params)
                lines!(ax2, lift(c -> [s.proposalParams.params[i] for s in c], dtchain))
            end
        end
        on(s -> autolimits!(ax2), dtchain)

        ax3 = Axis(infogrid[1,2], subtitle = "acceptance rate")
        acceptancerates = lift(NPhotons.acceptanceRate, dtchain)
        lines!(ax3, acceptancerates)
        scatter!(ax3, [1,1], [0, 1], color = :transparent)
        on(s -> autolimits!(ax3), dtchain)

        Label(infogrid[2,:], lift(c -> "T = $(round(c[end].temperature, sigdigits = 5))   E = $(round(c[end].energy, sigdigits = 5))   d = $(round.(last(c[end].proposalParams), sigdigits = 5))   nsteps = $(c[end].nsteps)", dtchain), tellwidth = false)
    end

    stepslider = Slider(fig[end+1, :], range = range(0, 1, length = 10000))
    on(liftevery(stepslider.value, 0.03)) do v dtchain[] = chain[][1:max(1, round(Int, end * v))] end
    set_close_to!(stepslider, 1)

    fig
end

NPhotons.monitor(chain, args...; kwargs...) = NPhotons.monitor(Observable(chain), args...; kwargs...)

end