import Base.length, Base.getindex

abstract type Structure end

struct AtomVolume{T, A <: AbstractVector{<:SVector{3, <:AbstractFloat}}, B <: AbstractVector{<:AbstractFloat}} <: Structure
    positions::A
    heights::B
    widths::B

    function AtomVolume(positions, heights, widths)
        new{eltype(eltype(positions)), typeof(positions), typeof(widths)}(positions, heights, widths)
    end
end

function AtomVolume(positions, heights, widths, radius)
    AtomVolume(positions, heights, widths)
end

function Base.zero(::AtomVolume{T, A, B}) where {T, A, B}
    AtomVolume(A(), B(), B())
end

function emptyAtomVolume()
    AtomVolume(SVector{3,Float64}[], Float64[], Float64[])
end

function AtomVolume(N::Int, σ::Real)
    AtomVolume(fill(SA[0.0, 0.0, 0.0], N), σ)
end

function AtomVolume(positions::AbstractVector{<:AbstractVector}, σ::Real)
    numtype = eltype(eltype(positions))
    AtomVolume(convert.(SVector{3, numtype}, positions), ones(length(positions)), fill(Float64(σ), length(positions)))
end

function Base.repeat(f::AtomVolume; kwargs...)
    AtomVolume(repeat(f.positions; kwargs...), repeat(f.heights; kwargs...), repeat(f.widths; kwargs...))    
end

function randomize(f::AtomVolume)
    s = norm(std(f.positions)) / 1.59
    newf = deepcopy(f)
    newf.positions .= s .* randn.(eltype(newf.positions))
    newf
end

function randomAtomVolume(N::Int, σ::Real, radius; sum = 1.0)
    positions = [SVector{3}(randfromSphere(3)) * radius for i in 1:N]
    positions .-= Ref(mean(positions))
    AtomVolume(positions, σ) * (sum / N)
end

getpositions(f::AtomVolume) = f.positions
getheights(f::AtomVolume) = f.heights
getwidths(f::AtomVolume) = f.widths

function Base.:(==)(f::AtomVolume, g::AtomVolume)
    f.positions == g.positions && f.heights == g.heights && f.widths == g.widths
end

function Base.isequal(f::AtomVolume, g::AtomVolume)
    isequal(f.positions, g.positions) && isequal(f.heights, g.heights) && isequal(f.widths, g.widths)
end

function Base.hash(f::AtomVolume, h::UInt)
    hash(f.positions, hash(f.heights, hash(f.widths, hash(:AtomVolume, h))))
end

function changePrecision(f::AtomVolume, T::Type)
    AtomVolume(SVector{3, T}.(f.positions), T.(f.heights), T.(f.widths))
end

function cuda(f::AtomVolume)
    cupositions = cu(SVector{3,Float32}.(f.positions))
    cuheights = cu(f.heights)
    cuwidths = cu(f.widths)
    cuf = AtomVolume(cupositions, cuheights, cuwidths)
end

function CUDA.Adapt.adapt_structure(to::CUDA.CuArrayKernelAdaptor, f::AtomVolume)
    f.positions isa CuArray && return f
    positions = CUDA.Adapt.adapt_structure(to, SVector{3, Float32}.(f.positions))
    heights = CUDA.Adapt.adapt_structure(to, f.heights)
    widths = CUDA.Adapt.adapt_structure(to, f.widths)
    AtomVolume(positions, heights, widths)
end

function CUDA.Adapt.adapt_structure(to, f::AtomVolume)
    positions = CUDA.Adapt.adapt_structure(to, f.positions)
    heights = CUDA.Adapt.adapt_structure(to, f.heights)
    widths = CUDA.Adapt.adapt_structure(to, f.widths)
    AtomVolume(positions, heights, widths)
end

function unsafe_free!(f::AtomVolume)
    f.positions isa CuArray && CUDA.unsafe_free!(f.positions)
    f.heights isa CuArray && CUDA.unsafe_free!(f.heights)
    f.widths isa CuArray && CUDA.unsafe_free!(f.widths)
end

function getindex(f::AtomVolume, i::Int)
    f.positions[i]
end

function getindex(f::AtomVolume, what::Symbol)
    x -> f(x, what)
end

function evaluateAtomVolume(positions, heights::AbstractArray{T}, widths, x)::T where T
    result = zero(Complex{T})
    dotxx = dot(x, x)
    if all(w == widths[1] for w in widths)
        for (p, h) in zip(positions, heights)
            result += h * cis(dot(x, p))
        end
        result *= exp(-widths[1]^2 * dotxx / 2)
    else
        for (p, h, σ) in zip(positions, heights, widths)
            result += @fastmath h * exp(complex(-σ^2 * dotxx / 2, dot(x, p)))
        end
    end
    abs2(result)
end

function ftgauss(σ, x, p)
    exp(complex(-σ^2 * dot(x, x) / 2, dot(x, p)))
end

function evaluateAtomVolumeComplex(positions, heights::AbstractArray{T}, widths, x)::Complex{T} where T
    result = zero(Complex{T})
    dotxx = dot(x, x)
    for (p, h, σ) in zip(positions, heights, widths)
        result += h * exp(complex(-σ^2 * dotxx / 2, dot(x, p)))
    end
    result
end

function toMatrix(f::AtomVolume, X)
    ftgauss.(f.widths, reshape(X, 1, size(X)...), f.positions)
end

function evaluate(f::AtomVolume, P)
    evaluateAtomVolume.(Ref(f.positions), Ref(f.heights), Ref(f.widths), P)
end

function evaluate!(X, f::AtomVolume, P)
    X .= evaluateAtomVolume.(Ref(f.positions), Ref(f.heights), Ref(f.widths), P)
end

function evaluate_complex(f::AtomVolume, P)
    evaluateAtomVolumeComplex.(Ref(f.positions), Ref(f.heights), Ref(f.widths), P)
end

function evaluate_complex!(X, f::AtomVolume, P)
    X .= evaluateAtomVolumeComplex.(Ref(f.positions), Ref(f.heights), Ref(f.widths), P)
end

function (f::AtomVolume)(x, y, z, args...; kwargs...)
    f(SA[x, y, z], args...; kwargs...)
end

function (f::AtomVolume)(x)
    evaluateAtomVolume(f.positions, f.heights, f.widths, x)
end

function (f::AtomVolume)(x, what; cutoff = Inf)
    if what == :real
        result = 0.0
        if cutoff == 0
            for (p, h, σ) in zip(f.positions, f.heights, f.widths)
                result += h * σ^-3 * exp(-1 / (2σ^2) * sqeuclidean(x, p))
            end
        else
            for (p, h, σ) in zip(f.positions, f.heights, f.widths)
                b = 1 / (2σ^2) * sqeuclidean(x, p)
                if b < cutoff
                    result += h * σ^-3 * (exp(-b) - exp(-cutoff))
                end
            end
        end
        return result / (2pi)^(3 / 2)
    elseif what == :fourier
        return f(x)
    elseif what == :complex 
        result = zero(Complex{eltype(f.heights)})
        dotxx = dot(x, x)
        for (p, h, σ) in zip(f.positions, f.heights, f.widths)
            result += h * exp(complex(-σ^2 * dotxx / 2, dot(x, p)))
        end
        result
    else
        zero(eltype(f.heights)) / zero(eltype(f.heights))
    end
end

function realspace(f::AtomVolume; cutoff = Inf)
    x -> f(x, :real; cutoff = Inf)
end

function evaluateRealFast(f, grid; cutoff = 10, showprogress = true)
    tree = KDTree(f.positions)
    prog = Progress(length(grid), enabled = showprogress)
    map(grid) do x
        next!(prog)
        sum(inrange(tree, x, cutoff); init = 0.0) do i
            f.heights[i] * f.widths[i]^-3 * exp(-1 / (2f.widths[i]^2) * sqeuclidean(x, f.positions[i]))
        end
    end
end

function gradient(f::AtomVolume, i, k)
    dotkk = dot(k, k)

    yi = f.positions[i]; hi = f.heights[i]; σi = f.widths[i]

    result = zero(Complex{eltype(f.heights)})
    if false#all(w == f.widths[1] for w in f.widths)
        for (yj, hj) in zip(f.positions, f.heights)
            result += hj * cis(dot(k, yi - yj))
        end
        result *= exp(-f.widths[1]^2 * dotkk)
    else
        for (yj, hj, σj) in zip(f.positions, f.heights, f.widths)
            result += hj * exp(complex(-(σi^2+σj^2) * dotkk / 2, dot(k, yi - yj)))
        end
    end
    k * real(2im * hi * result)
end

function l1(f::AtomVolume)
    sum(f.heights)
end

function LinearAlgebra.normalize(f::AtomVolume)
    f / l1(f)
end

function LinearAlgebra.normalize!(f::AtomVolume)
    f.heights ./= l1(f)
    f
end

function scatteringProbability(f::AtomVolume; λ = 2.5, qmax = Inf)
    result = 0.0
    r = min(2pi / λ, qmax / 2)
    for (p1, h1, σ1) in zip(f.positions, f.heights, f.widths)
        for (p2, h2, σ2) in zip(f.positions, f.heights, f.widths)
            k = norm(p1 - p2)
            s = sqrt(σ1^2 + σ2^2)
            if k <= 1e-5
                result += h1 * h2 * (1 - exp(-2*r^2*s^2)) / s^2
            else
                x = 2dawson(k / (√2 * s))
                x -= dawson((k - 2im * s^2 * r) / (√2 * s)) * exp(((k - 2im * s^2 * r)^2 - k^2) / (2s^2))
                x -= dawson((k + 2im * s^2 * r) / (√2 * s)) * exp(((k + 2im * s^2 * r)^2 - k^2) / (2s^2))
                result += real(h1 * h2 * x * sqrt(2) / (2k * s))
            end
        end
    end
    result * 2pi
end

function normalizeScattering(f::AtomVolume; nphotons = 1, λ = 2.5, qmax = Inf)
    f * sqrt(nphotons / scatteringProbability(f, λ = λ, qmax = qmax))
end


function envelope(f::AtomVolume)
    o = SA[0.0,0.0,0.0]
    e = AtomVolume(SA[o], SA[1.0], SA[minimum(f.widths)])
    envf = @set e.heights = SA[sqrt(f(o) / e(o))]
    r -> envf(SA[r,0.0,0.0])
end

function normalEnvelope(f::AtomVolume; dims = 3)
    MvNormal(fill(0.0, dims), 1 / (√2 * minimum(f.widths)))
end

function upperBound(f::AtomVolume, r)
    result = 0.0
    a = r^2 / 2
    for (h, σ) in zip(f.heights, f.widths)
        result += h * exp(-σ^2 * a)
    end
    abs2(result)
end

function smoothen(f::AtomVolume, σ)
    newf = deepcopy(f)
    newf.widths .= sqrt.(σ^2 .+ f.widths.^2);
    newf
end

function sharpen(f::AtomVolume, σ)
    newf = deepcopy(f)
    newf.widths .= sqrt.(f.widths.^2 .- σ^2);
    newf
end

function length(f::AtomVolume)
    return length(f.positions)
end

Base.broadcastable(f::AtomVolume) = Ref(f)

function Base.show(io::IO, f::AtomVolume)
    compact = get(io, :compact, false)

    println(io, "$(length(f))-element $(typeof(f)):")
    println(io, "    Positions: ", string((p -> round.(p, digits = 3)).(f.positions[1:min(2,end)]))[1:end-1], ", ... ]")
    println(io, "    Heights: ", string(round.(f.heights[1:(min(6,end))], digits = 3))[1:end-1], ", ... ]")
    print(io, "    Widths: ", string(round.(f.widths[1:(min(6,end))], digits = 3))[1:end-1], ", ... ]")
end

function Base.:+(f::AtomVolume, g::AtomVolume)
    AtomVolume(vcat(f.positions, g.positions), vcat(f.heights, g.heights), vcat(f.widths, g.widths))
end

function rotate(f::AtomVolume, R)
    newf = deepcopy(f)
    for i in eachindex(f.positions)
        newf.positions[i] = R * f.positions[i]
    end
    newf
end

Base.:*(R, f::AtomVolume) = rotate(f, R)
Base.:*(a::Real, f::AtomVolume) = (g = deepcopy(f); g.heights .*= a; g)
Base.:*(f::AtomVolume, a::Real) = a * f
Base.:/(f::AtomVolume, a::Real) = a^-1 * f

function align(f::AtomVolume, g::AtomVolume; ntries = 1000, mirror = 0.5, full = true, σ = 0.5, showprogress = false, rng = Random.GLOBAL_RNG)
    fpos = f.positions; gpos = g.positions
    fh = f.heights; gh = g.heights

    dists = zeros(length(fpos), length(gpos))
    dX = zeros(length(fpos))
    dY = zeros(length(gpos))

    R = SMatrix{3, 3, Float64, 9}(I)
    Rfpos = Ref(R) .* fpos
    dR = pointCloudDistance!(dists, dX, dY, (Rfpos .= Ref(R) .* fpos), gpos, fh, gh)

    prog = Progress((1 + full) * ntries)
    
    if full
        for i in 1:ntries
            S = SMatrix{3, 3, Float64, 9}(randomRotation(rng))
            if rand(rng) < mirror
                S *= -1
            end
            dS = pointCloudDistance!(dists, dX, dY, (Rfpos .= Ref(S) .* fpos), gpos, fh, gh)
            if dS < dR
                R = S
                dR = dS
            end
            showprogress && next!(prog)
        end
    end
    
    for i in 1:ntries
        S = randomSmallRotation(σ, rng)
        dS = pointCloudDistance!(dists, dX, dY, (Rfpos .= Ref(R * S) .* fpos), gpos, fh, gh)
        if dS < dR
            R = R * S
            dR = dS
        end
        showprogress && next!(prog)
    end
    
    R
end

function aligned(f::AtomVolume, g::AtomVolume; kwargs...)
    newf = deepcopy(f)
    newf.positions .-= Ref(mean(newf.positions, Weights(newf.heights)))
    align(newf, g; kwargs...) * newf
end

function alignReal(f::AtomVolume, g::AtomVolume; kwargs...)
    radius = 1.2maximum(maximum.(norm, (f.positions, g.positions)))
    align(realspace(f), realspace(g); radius = radius, kwargs...)
end

function alignedReal(f::AtomVolume, g::AtomVolume; kwargs...)
    newf = deepcopy(f)
    newf.positions .-= Ref(mean(newf.positions, Weights(newf.heights)))
    alignReal(newf, g; kwargs...) * newf
end