struct IntegrationParameters
    rotations::Vector{QuatRotation{Float64}}
    rotationWeights::Vector{Float64}
    
    pixels::Vector{SVector{3, Float64}}
    pixelWeights::Vector{Float64}

    precision::Int
    qstep::Float64
    qmax::Float64
    npercircle::Int
    wavelength::Float64
end

function IntegrationParameters(; precision = 23, qstep, qmax, npercircle = 32, wavelength = 2.5)
    pixels, pixelWeights = weightedEwaldGrid(qstep = qstep, qmax = qmax, npercircle = npercircle, wavelength = wavelength)
    rotations, rotationWeights = rotationGrid(precision = precision)
    IntegrationParameters(rotations, rotationWeights, pixels, pixelWeights, precision, qstep, qmax, npercircle, wavelength)
end

function Base.show(io::IO, params::IntegrationParameters)
    show(io, IntegrationParameters)
    print(io, "(precision = $(params.precision), qstep = $(params.qstep), qmax = $(params.qmax), npercircle = $(params.npercircle), wavelength = $(params.wavelength))")
end


struct ComputeConfiguration
    params::IntegrationParameters
    indexImages::Vector{Vector{Int}}

    ncalls::Int
    inner::Bool
    nhitvalues::Int
end

function ComputeConfiguration(images; qmax, σreg = 0.0, nimages = Inf, ncalls = 10, inner = true, nhitvalues = 10, sortimages = true, showprogress = true, kwargs...)
    params = IntegrationParameters(; qmax, kwargs...)
    indexImages = if eltype(eltype(images)) <: Integer
        copy(images)
    else
        loadIndexImages(images, params.pixels; σ = σreg, qmax = qmax, nimages = nimages, showprogress = showprogress)
    end
    sortimages && sort!(indexImages, by = length)
    ComputeConfiguration(params, indexImages, ncalls, inner, nhitvalues)
end

function distribute(conf::ComputeConfiguration, workerweights::AbstractArray)
    assignments = assignJobs(length(conf.indexImages), workerweights)
    [@set conf.indexImages = conf.indexImages[a] for a in assignments]
end
distribute(conf::ComputeConfiguration, n::Int) = distribute(conf, ones(n))

Base.length(conf::ComputeConfiguration) = length(conf.indexImages)

function Base.show(io::IO, conf::ComputeConfiguration)
    show(io, ComputeConfiguration)
    print(io, "(ncalls = $(conf.ncalls), inner = $(conf.inner), nhitvalues = $(conf.nhitvalues))")

    if !get(io, :compact, false)
        params = conf.params
        println(io, ":")
        println(io, "  precision = $(params.precision), qstep = $(params.qstep), qmax = $(params.qmax), npercircle = $(params.npercircle), wavelength = $(params.wavelength)")
        print(io, "  $(length(conf.indexImages)) images, ")
        print(io, "$(sum(length, conf.indexImages)) photons, ")
        print(io, "$(length(conf.params.rotations)) × $(conf.params.npercircle) = $(length(conf.params.rotations)) rotations, ")
        print(io, "$(length(conf.params.pixels)÷conf.params.npercircle) × $(conf.params.npercircle) = $(length(conf.params.pixels)) pixels")
    end
end

struct ComputeAllocation{M, I, UI, F1, F1D, F2, F2D, F3, V1, V2, P, D, NF2, B}
    configuration::ComputeConfiguration
    forwardmodel::M

    imageLocations::I
    imageLengths::I
    indexImages::UI

    fvalues::F2
    fvaluesDuplicated::F2
    meanfvalues::F1
    
    pixels::V1
    rotatedPixels::V2
    pixelWeights::F1
    rotationWeights::F1

    nphotonsPerRotation::F2
    nphotonLikelihoods::F3
    maxnphotonLikelihoods::F1
    
    polarizationFactors::P
    polarizationFactorsDuplicated::P

    detectionFactors::D
    weightedDetectionFactors::D

    premulPixelWeights::NF2

    bvalues::B
    bvaluesDuplicated::B
    
    scratch1::F1D
    scratch2::F1

    kernelResults::F2D
    results::F1D
end

macro unpack_Allocation(q)
    code = Expr(:block, [ :($field = $q.$field) for field in fieldnames(ComputeAllocation) ]...)
    esc(code)
end

function ComputeAllocation(::Type{T}, ::Type{N}, ::Type{ND}, model::AbstractForwardModel, conf::ComputeConfiguration) where {T, N, ND}
    oversampledetector = 10
    npercircle = conf.params.npercircle

    imageLengths = UInt32.(length.(conf.indexImages))
    shiftedIndexImages = [image .+ convert.(eltype(image), npercircle .* ((image .- 1) .÷ npercircle)) for image in conf.indexImages]
    imageLocations, serializedIndexImages = serializeIndexImages(shiftedIndexImages)
    serializedIndexImages = convert(Vector{UInt16}, serializedIndexImages)
    imageLocations = T(UInt32.(imageLocations)); imageLengths = T(imageLengths); serializedIndexImages = T(serializedIndexImages)
    

    npix = length(conf.params.pixels); nrot = length(conf.params.rotations); 
    nimg = length(conf.indexImages); lmax = maximum(imageLengths, init = 0)


    fvalues = T{N}(undef, npix, nrot)
    fvaluesDuplicated = T{N}(undef, 2npix, nrot)
    meanfvalues = T{N}(undef, 2npix + 1); meanfvalues .= 1

    pixels = T{SVector{3, Float32}, 1}(conf.params.pixels)
    rotatedPixels = T{SVector{3, Float32}, 2}([R * x for x in conf.params.pixels, R in conf.params.rotations])
    pixelWeights = T{N}(conf.params.pixelWeights)
    rotationWeights = T{N}(conf.params.rotationWeights)

    premulPixelWeights = if model.polarization || !isa(model.detector, IdealDetector)
        arr = zeros(length(pixelWeights), npercircle)
        for i in 0:npercircle-1
            circshift!(view(reshape(arr, npercircle, :, npercircle),:,:,i+1), reshape(conf.params.pixelWeights, npercircle, :), (i, 0))
        end
        arr
    else
        nothing
    end

    polarizationFactors = model.polarization ? polarizationFactor.(conf.params.pixels; λ = model.λ) : nothing
    polarizationFactorsDuplicated = if model.polarization
        rpolfacs = reshape(polarizationFactors, npercircle, :)
        T{N}(vcat(rpolfacs, rpolfacs)[:])
    else
        nothing
    end

    detectionFactors = if !isa(model.detector, IdealDetector)
        ps, ws = weightedEwaldGrid(
            qstep = conf.params.qstep / oversampledetector, qmax = conf.params.qmax, 
            npercircle = oversampledetector * npercircle, wavelength = model.λ
        )
        ds = zeros(length(conf.params.pixels))
        counts = zeros(Int, length(conf.params.pixels))
        for (i, p) in zip(nearestneighbors(conf.params.pixels, ps), ps)
            ds[i] += model.detector(p)
            counts[i] += 1
        end
        ds ./= max.(1, counts)
        ds
    else
        nothing
    end

    if model.polarization
        for i in 0:npercircle-1
            view(reshape(premulPixelWeights, npercircle, :, npercircle),:,:,i+1) .*= circshift(reshape(polarizationFactors, npercircle, :), (i, 0))
        end
    end
    if !isa(model.detector, IdealDetector)
        for i in 0:npercircle-1
            view(reshape(premulPixelWeights, npercircle, :, npercircle),:,:,i+1) .*= circshift(reshape(detectionFactors, npercircle, :), (i, 0))
        end
    end


    polarizationFactors = isnothing(polarizationFactors) ? nothing : T{N}(polarizationFactors)
    detectionFactors = isnothing(detectionFactors) ? nothing : T{N}(detectionFactors)
    weightedDetectionFactors = isnothing(detectionFactors) ? nothing : detectionFactors .* pixelWeights
    premulPixelWeights = isnothing(premulPixelWeights) ? nothing : T{N}(premulPixelWeights)

    nphotonsPerRotation = T{N}(undef, nrot, isnothing(premulPixelWeights) ? 1 : npercircle)
    nphotonLikelihoods = T{ND}(undef, lmax + 1, nrot, (isnothing(premulPixelWeights) ? () : (npercircle,))...)
    maxnphotonLikelihoods = T{N}(undef, lmax + 1)

    bvalues = hasnoise(model) ? T{N}(undef, npix) : nothing
    bvaluesDuplicated = hasnoise(model) ? T{N}(undef, 2npix) : nothing

    scratch1 = T{ND}(undef, nimg)
    scratch2 = T{N}(undef, 2npix ÷ npercircle)

    kernelResults = T{ND}(undef, nimg, ceil(Int, nrot / conf.ncalls))
    results = T{ND}(undef, nimg)

    ComputeAllocation(conf, model, imageLocations, imageLengths, serializedIndexImages, fvalues, fvaluesDuplicated, meanfvalues, pixels, rotatedPixels, pixelWeights, rotationWeights, nphotonsPerRotation, nphotonLikelihoods, maxnphotonLikelihoods, polarizationFactors, polarizationFactorsDuplicated, detectionFactors, weightedDetectionFactors, premulPixelWeights, bvalues, bvaluesDuplicated, scratch1, scratch2, kernelResults, results)
end

function ComputeAllocation(model::AbstractForwardModel, conf::ComputeConfiguration)
    ComputeAllocation(CuArray, Float32, model, conf)
end

function ComputeAllocation(T1, T2, model::AbstractForwardModel, conf::ComputeConfiguration)
    ComputeAllocation(T1, T2, T2, model, conf)
end

function ComputeAllocation(T1, T2s::Tuple, model::AbstractForwardModel, conf::ComputeConfiguration)
    ComputeAllocation(T1, T2s[1], T2s[2], model, conf)
end

function iscuda(alloc::ComputeAllocation)
    alloc.fvalues isa CuArray
end

function minsize!(alloc::ComputeAllocation, n)
    if length(alloc.results) < length(alloc.imageLocations) * n
        resize!(alloc.results, length(alloc.imageLocations) * n)
    end
end

function compute_foreground!(alloc::ComputeAllocation, f, η = 1)
    cf = iscuda(alloc) ? cu(f) : f
    alloc.fvalues .= η .* cf.(alloc.rotatedPixels)
    cf !== f && unsafe_free!(cf)
end

function compute_background!(alloc::ComputeAllocation, m, η = 1)
    C(x) = iscuda(alloc) ? cu(x) : x

    cf_inel = C(m.inelastic)
    cb_el = C(m.background_elastic)
    cb_inel = C(m.background_inelastic)

    if !isnothing(m.background_elastic)
        alloc.bvalues .= alloc.polarizationFactors .* cb_el.(alloc.pixels) # fix no polarization when noise
        alloc.bvalues .+= η .* cf_inel.(alloc.pixels)
        alloc.bvalues .+= cb_inel.(alloc.pixels)
    end

    m.inelastic !== cf_inel && unsafe_free!(cf_inel); 
    m.background_elastic !== cb_el && unsafe_free!(cb_el); 
    m.background_inelastic !== cb_inel && unsafe_free!(cb_inel)
end

function compute_intensities!(alloc::ComputeAllocation, f, m, η = 1)
    compute_foreground!(alloc, f, η)
    compute_background!(alloc, m, η)
end


function applyKernel!(alloc::ComputeAllocation, rotationrange)
    applyKernel!(alloc.kernelResults, alloc.fvaluesDuplicated, alloc.nphotonLikelihoods, alloc.imageLocations, alloc.indexImages, alloc.configuration.params.npercircle - 1, rotationrange, alloc.polarizationFactorsDuplicated, alloc.bvaluesDuplicated)
end

function compute_imagelikelihoods!(alloc::ComputeAllocation, f, model::ForwardModel; results_slot = 1, η = 1)
    minsize!(alloc, results_slot)

    @unpack_Allocation(alloc)
    npercircle = configuration.params.npercircle

    duplicateChunks!(fvaluesDuplicated, fvalues, npercircle)
    !isnothing(bvalues) && duplicateChunks!(bvaluesDuplicated, bvalues, npercircle)
    
    mul!(nphotonsPerRotation, fvalues', isnothing(premulPixelWeights) ? pixelWeights : premulPixelWeights)

    if !isnothing(bvalues)
        if !isnothing(weightedDetectionFactors)
            nphotonsPerRotation .+= bvalues ⋅ weightedDetectionFactors
        else
            nphotonsPerRotation .+= bvalues ⋅ pixelWeights
        end
    end

    rnPR = reshape(nphotonsPerRotation, 1, size(nphotonsPerRotation)...)

    if model.intensitydist isa Tuple{Real}
        nphotonLikelihoods .= .- (model.meanintensity * mean(model.intensitydist)) .* rnPR
    elseif model.intensitydist isa Gamma
        α = model.intensitydist.α
        β = 1 / model.intensitydist.θ / model.meanintensity
        nphotonLikelihoods .= log.(rnPR .+ β) .* ((0: -1 : -(size(nphotonLikelihoods, 1) - 1)) .- α)
    elseif isnothing(model.intensitydist)
        nphotonLikelihoods .= log.(rnPR) .* (0: -1 : -(size(nphotonLikelihoods, 1) - 1))
    end

    # CUDA.@time maximum!(maxnphotonLikelihoods, nphotonLikelihoods)
    maxnphotonLikelihoods .= @view nphotonLikelihoods[:, 1, 1]
    nphotonLikelihoods .= exp.(nphotonLikelihoods .- maxnphotonLikelihoods)

    v = @view(meanfvalues[2:end])
    mean!(v, fvaluesDuplicated)
    !isnothing(bvalues) && (v .+= bvaluesDuplicated)
    reshape(v, npercircle, :) .= mean!(reshape(scratch2, 1, :), reshape(v, npercircle, :))
    fvaluesDuplicated ./= v; 
    !isnothing(bvalues) && (bvaluesDuplicated ./= v)
    
    results = view(results, (1:length(imageLocations)) .+ (results_slot - 1) * length(imageLocations))
    results .= 0
    
    for rotationRange in partitionrange(1:length(rotationWeights), stride = size(kernelResults, 2))
        kernelResults .= 0
        applyKernel!(alloc, rotationRange)
        view(kernelResults, :, 1:length(rotationRange)) .*= reshape(view(rotationWeights, rotationRange), 1, :)
        sum!(scratch1, kernelResults); results .+= scratch1
    end

    results .= log.(results ./ (configuration.inner ? npercircle : 1))
    broadcast!(results, results, Ref(meanfvalues), Ref(indexImages), imageLocations) do r, meanfvalues, indexImages, ind
        x = zero(r)
        for k in 1:indexImages[ind]
            x += log(meanfvalues[indexImages[ind + k] + 1])
        end
        r + x
    end
    results .+= getindex.(Ref(maxnphotonLikelihoods), imageLengths .+ 1)
    results
end

function compute_imagelikelihoods!(alloc::ComputeAllocation, ensemble::WeightedEnsemble, model::ForwardModel; η = 1)
    for (i, f) in enumerate(ensemble.structs)
        compute_imagelikelihoods!(alloc, f, model; results_slot = i, kwargs...)
    end
end

function compute_imagelikelihoods!(alloc::ComputeAllocation, f, model::ForwardModel, ηs::AbstractArray)
    for (i, η) in enumerate(ηs)
        compute_imagelikelihoods!(alloc, f, model; results_slot = i, η = η)
    end
end

function logsumexp(alloc::ComputeAllocation, weights::AbstractArray)
    minsize!(alloc, 2length(weights) + 1)

    nimg = length(alloc.imageLocations); lens = length(weights)
    results = reshape(@view(alloc.results[1 : nimg * lens]), nimg, lens)

    A = reshape(@view(alloc.results[nimg * lens + 1 : 2nimg * lens]), nimg, lens)
    B = @view(alloc.results[2nimg * lens + 1 : end])
    M = alloc.scratch1; 
    W = typeof(results)(normalize(weights, 1)')

    logsumexp!(M, A, B, W, results)
    
    W isa CuArray && CUDA.unsafe_free!(W)
    
    sum(M, init = 0.0)
end

function (alloc::ComputeAllocation)(f::Structure, model::ForwardModel{<:Any, <:Any, <:Any}; recompute = true)
    compute_intensities!(alloc, f, model)
    sum(compute_imagelikelihoods!(alloc, f, model), init = 0.0)
end

function (alloc::ComputeAllocation)(f::Structure, model::ForwardModel{<:Any, <:Any, <:Distribution}; recompute = true)
    minsize!(alloc, 2alloc.configuration.nhitvalues + 1)
    
    ηs = range(extrema(model.hitdist)..., length = alloc.configuration.nhitvalues)
    cdfs = [0; cdf.(model.hitdist, middle.(ηs[2:end], ηs[1:end-1])); 1]
    ws = cdfs[2:end] .- cdfs[1:end-1]

    if recompute
        cf = iscuda(alloc) ? cu(f) : f
        cm = iscuda(alloc) ? cuda(model) : model
        for (i, η) in enumerate(ηs)
            compute_intensities!(alloc, cf, cm, η)
            compute_imagelikelihoods!(alloc, f, model; results_slot = i)
        end
        unsafe_free!(cf)
        unsafe_free!(cm)
    end

    logsumexp(alloc, ws)
end

function (alloc::ComputeAllocation)(ensemble::WeightedEnsemble, model::ForwardModel; recompute = true)
    minsize!(alloc, 2length(ensemble) + 1)

    if recompute
        compute_background!(alloc, model)
        for (i, f) in enumerate(ensemble.structs)
            compute_foreground!(alloc, f)
            compute_imagelikelihoods!(alloc, f, model; results_slot = i)
        end
    end

    logsumexp(alloc, ensemble.weights)
end

function (alloc::ComputeAllocation)(f; kwargs...)
    alloc(f, alloc.forwardmodel; kwargs...)
end