
function fsc(f1::AtomVolume, f2::AtomVolume, radius::Real, points, weights)
    fsc(f1[:complex], f2[:complex], radius, points, weights)
end

function fsc(y1::AbstractArray, y2::AbstractArray, weights::AbstractArray)
    A = Diagonal(weights)
    real(dot(y1, A, y2) / sqrt(dot(y1, A, y1) * dot(y2, A, y2)))
end

function fsc(f1, f2, radius::Real, points, weights)
    fsc(f1.(radius .* points), f2.(radius .* points), weights)
end

function fsc(f1, f2, radius::Real; precision = 41)
    points, weights = lebedevGrid(precision = precision)
    fsc(f1, f2, radius, points, weights)
end

function fsc(f1, f2, radii; precision = 41)
    points, weights = lebedevGrid(precision = precision)
    [fsc(f1, f2, r, points, weights) for r in radii]
end

function fscResolution(f1, f2; align = false, ntries = 1000, kwargs...)
    align && (f1 = aligned(f1, f2, ntries = ntries))
    2pi / fscResolutionFourier(f1, f2; kwargs...)
end

function fscResolution(f1, f2, points, weights; kwargs...)
    2pi / fscResolutionFourier(f1, f2, points, weights; kwargs...)
end

function fscResolutionFourier(f1, f2; precision = 41, kwargs...)
    points, weights = lebedevGrid(precision = precision)
    fscResolutionFourier(f1, f2, points, weights; kwargs...)
end

function fscResolutionFourier(f1, f2, points, weights; rstep = 0.1, threshold = 0.5, rmax = 100.0, tol = 1e-5)
    r = 0.0
    for i in 1:100000
        if fsc(f1, f2, r + rstep, points, weights) > threshold
            r += rstep
        else
            rstep /= 2
        end
        if rstep < tol || r >= rmax
            return r
        end
    end
    r
end

function alignFSC_cd(f1, f2; rstep = 0.01, threshold = 0.5, precisionFSC = 23)
    points, weights = lebedevGrid(precision = precisionFSC)
    R = I
    res = fscResolutionFourier(f1, f2, points, weights, rstep = rstep)
    astep = 0.5
    @showprogress for i in 1:100
        for (j, axis) in enumerate([[1, 0, 0], [0, 1, 0], [0, 0, 1], [-1, 0, 0], [0, -1, 0], [0, 0, -1]])
            newres = fscResolutionFourier(f1 ∘ Base.Fix1(*, AngleAxis(astep, axis...) * R), f2, points, weights, rstep = rstep)
            if newres >= res
                res = newres
                R = AngleAxis(astep, axis...) * R
                # break
            end
            # if j == 6 return R end
        end
        astep *= 0.9
    end
    return R
end

function alignFSC_rd(f1, f2; rstep = 0.01, threshold = 0.5, precision = 23, astep = pi, tol = 1e-3, alambda = 1.01, nsteps = 1000, showprogress = true)
    points, weights = lebedevGrid(precision = precision)
    R = SMatrix{3, 3, Float64, 9}(I)
    res = fscResolutionFourier(f1, f2, points, weights, rstep = rstep)
    asteps = Float64[]
    ress = Float64[]
    prog = Progress(nsteps, enabled = showprogress)
    for i in 1:nsteps
        Q = randomSmallRotation(astep) * rand((-1,1))
        # Q = AngleAxis(astep, normalize(randn(3))...)
        newres = fscResolutionFourier(Q * R * f1, f2, points, weights, rstep = rstep, tol = tol)
        if newres > res
            res = newres
            R = Q * R
            astep *= alambda
        else
            astep /= alambda
        end
        push!(asteps, astep)
        push!(ress, res)
        next!(prog, showvalues = ((:astep, astep), (:res, 2pi/res)))
        astep < 0.01 && break
    end
    return R#, asteps, ress
end


function alignFSC(f1, f2; rstep = 0.01, threshold = 0.5, precisionFSC = 23, precisionRot = 23, maxr = 100.0, npercircle = 64)
    points, weights = lebedevGrid(precision = precisionFSC)
    rotations = uniformRotationGrid(precision = precisionRot, npercircle = npercircle)[1]
    resolutions = @showprogress [fscResolutionFourier(f1 ∘ Base.Fix1(*, R), f2, points, weights, rstep = rstep) for R in rotations]
    rotations[argmax(resolutions)]
end

function alignFSC(f1::AtomVolume, f2::AtomVolume; kwargs...)
    alignFSC(f1[:complex], f2[:complex]; kwargs...)
end

function alignFSCsel(f1, f2; rstep = 0.01, nrot = 1000, precision = 23, rmax = 100, showprogress = true)
    Rs = [randomRotation() for _ in 1:nrot]
    # Rs = uniformRotationGrid(precision = precision, npercircle = 64)[1][:]

    prog = ProgressUnknown(enabled = showprogress)
    for r in rstep:rstep:rmax
        fscs = [fscAnalytic(R * f1, f2, r) for R in Rs]
        if all(fscs .< 0.5)
            return Rs[1]
        else
            Rs = Rs[fscs .>= 0.5]
        end
        next!(prog, showvalues = ((:nrot, length(Rs)), (:r, r)))
    end
end

function alignedFSCsel(f1, f2; kwargs...)
    R = alignFSCsel(f1, f2; kwargs...)
    R * f1
end

function fscResolutionAnalytic(f1, f2; align = false, ntries = 1000, kwargs...)
    align && (f1 = aligned(f1, f2, ntries = ntries))
    2pi / fscResolutionFourierAnalytic(f1, f2; kwargs...)
end

function fscResolutionFourierAnalytic(f1, f2; rstep = 0.1, threshold = 0.5, rmax = 100.0, tol = 1e-5)
    r = 0.0
    for i in 1:100000
        if fscAnalytic(f1, f2, r + rstep) > threshold
            r += rstep
        else
            rstep /= 2
        end
        if rstep < tol || r >= rmax
            return r
        end
    end
    r
end


function radialCovariance(f1, f2, r, s = 0.0)
    result = 0.0
    for (p1, h1, σ1) in zip(f1.positions, f1.heights, f1.widths)
        for (p2, h2, σ2) in zip(f2.positions, f2.heights, f2.widths)
            result += h1 * h2 * sinc(r * norm(p1 - p2) / pi) * exp(-(σ1^2 + σ2^2 - 2s^2) * r^2 / 2)
        end
    end
    result
end

function radialCovariance(f1, f2, rs::AbstractArray, s = 0.0)
    b = broadcast(f1.positions, f1.heights, f1.widths, reshape(f2.positions, 1, :), f2.heights', f2.widths', reshape(rs, 1, 1, :)) do p1, h1, σ1, p2, h2, σ2, r
        h1 * h2 * sinc(r * norm(p1 - p2) / pi) * exp(-(σ1^2 + σ2^2 - 2s^2) * r^2 / 2)
    end
    reshape(sum(b, dims = (1,2)), :)
end

function fscAnalytic(f1, f2, r)
    s = min(minimum(f1.widths), minimum(f2.widths))
    radialCovariance(f1, f2, r, s) / sqrt(radialCovariance(f1, f1, r, s) * radialCovariance(f2, f2, r, s))
end

function fscAnalytic(f1, f2, rs::AbstractArray)
    s = min(minimum(f1.widths), minimum(f2.widths))
    radialCovariance(f1, f2, rs, s) ./ sqrt.(radialCovariance(f1, f1, rs, s) .* radialCovariance(f2, f2, rs, s))
end


function fscResolutionAnalyticCUDA(f1, f2; align = false, ntries = 1000, kwargs...)
    align && (f1 = aligned(f1, f2, ntries = ntries))
    2pi / fscResolutionFourierAnalyticCUDA(f1, f2; kwargs...)
end

function fscResolutionFourierAnalyticCUDA(f1, f2; rstep = 0.1, threshold = 0.5, rmax = 1.0)
    rs = 0.0f0:Float32(rstep):Float32(rmax)
    fscs = Array(fscAnalytic(cu(f1), cu(f2), rs))
    ind = findfirst(<(threshold), fscs)
    isnothing(ind) && return rmax
    Float64(rs[ind])
end

function alignFSCCUDA(f1, f2; nrot = 1000, kwargs...)
    Rs = [randomRotation() for _ in 1:nrot]
    res = fscResolutionFourierAnalyticCUDA.(Rs .* f1, f2; kwargs...)
    Rs[argmax(res)]
end

function alignedFSCCUDA(f1, f2; kwargs...)
    R = alignFSCCUDA(f1, f2; kwargs...)
    R * f1
end

function alignFSCmean(fs; nit = 10, ntries = 100, rmax = 1.0, ngrid = 10, showprogress = true, Rstep = 0.5, threshold = 0.5, mode = :threshold)
    xs = LinRange(-rmax, rmax, ngrid)
    grid = NPhotons.grid(xs, dims = 3)

    prog = Progress(nit*length(fs), enabled = showprogress)
    ress = zeros(length(fs))

    vals = @showprogress [f.(grid, :complex) for f in fs]
    itps = [splat(linear_interpolation((xs, xs, xs), v, extrapolation_bc = 0.0im)) for v in vals]

    Rs = [randomRotation()^0 for f in fs]
    try
        for _ in 1:nit
            for i in 1:length(fs)
                newRs = [[Rs[i]]; [(StatsBase.sample([-1,1], Weights([0.1, 0.9])) * randomSmallRotation(Rstep)) * Rs[i] for _ in 1:ntries]]

                c1(R) = v -> itps[i](R * v)
                c2(R) = v -> sum(itps[j]((i == j ? R : Rs[j]) * v) for j in 1:length(itps))
                if mode == :threshold
                    res = [fscResolution(c1(R), c2(R), rstep = 0.001, precision = 23, rmax = rmax, threshold = threshold) for R in newRs]
                elseif mode == :mean
                    res = [1 - mean(fsc(c1(R), c2(R), range(0, rmax, length = 25), precision = 23)) for R in newRs]
                end
                Rs[i] = newRs[argmin(res)]

                ress[i] = minimum(res)
                next!(prog, showvalues = ((:all, round.(ress, sigdigits = 3)), (:angle, round.(180 ./ pi .* rotation_angle.(Rs), sigdigits = 3)), (:rold, res[1]), (:rnew, minimum(res))))
            end
        end
    catch e
        if !isa(e, InterruptException)
            rethrow(e)
        end
    end
    Rs .^ -1
end

function alignFSCitp(f1, f2; ntries = 1000, rmax = 1.0, ngrid = 10, showprogress = true, Rstep = 0.5)
    xs = LinRange(-rmax, rmax, ngrid)
    grid = NPhotons.grid(xs, dims = 3)

    prog = Progress(ntries, enabled = showprogress)

    vals1 = @showprogress map(f1[:complex], grid)
    vals2 = @showprogress map(f2[:complex], grid)
    itp1 = splat(linear_interpolation((xs, xs, xs), vals1, extrapolation_bc = 0.0im))
    itp2 = splat(linear_interpolation((xs, xs, xs), vals2, extrapolation_bc = 0.0im))
    
    c(R) = v -> itp1(R * v)
    fscres(R) = fscResolution(c(R), itp2, rstep = 0.001, precision = 23, tol = 1e-5, rmax = rmax)

    R = randomRotation()^0
    res = fscres(R)

    try
        for _ in 1:ntries
            newR = StatsBase.sample([-1,1], Weights([0.1, 0.9])) * randomSmallRotation(Rstep) * R
            newres = fscres(newR)

            if newres <= res
                R = newR
                res = newres
            end
            next!(prog, showvalues = ((:res, res),))
        end
    catch e
        if !isa(e, InterruptException)
            rethrow(e)
        end
    end
    R^-1
end

function align_fscs_mean(fs; nit = 10, ntries = 100, rmax = 1.0, ngrid = 10, showprogress = true, Rstep = 0.5, threshold = 0.5)
    xs = LinRange(-rmax, rmax, ngrid)
    grid = NPhotons.grid(xs, dims = 3)

    prog = Progress(nit*length(fs), enabled = showprogress)

    vals = @showprogress [f.(grid, :complex) for f in fs]
    itps = [splat(linear_interpolation((xs, xs, xs), v, extrapolation_bc = 0.0im)) for v in vals]

    Rs = [randomRotation()^0 for f in fs]

    score(f, R, g, S) = mean(fsc(f ∘ Base.Fix1(*, R), g ∘ Base.Fix1(*, S), range(0, rmax, length = 25), precision = 23))

    scores = @showprogress [score(f, R, g, S) for (f, R) in zip(itps, Rs), (g, S) in zip(itps, Rs)]

    try
        for _ in 1:nit
            for i in 1:length(fs)
                newR = randomSmallRotation(Rstep) * Rs[i]

                newscores = map(1:length(fs), Rs, itps) do j, S, itp
                    if i != j
                        score(itps[1], newR, itp, S)
                    else
                        1.0
                    end
                end
                
                if mean(newscores) > mean(scores[:,i])
                    scores[:,i] .= newscores
                    scores[i,:] .= newscores
                    Rs[i] = newR
                end

                next!(prog, showvalues = ((:all, join(lpad.(round.(mean.(eachcol(scores)), sigdigits = 3),5), ", ")),))
            end
        end
    catch e
        if !isa(e, InterruptException)
            rethrow(e)
        end
    end

    Rs .^ -1
end