function kernel_gradient_1!(R, G, X, dX, P, I, ioff, ::Val{maxo}, ::Val{stride}) where {maxo, stride}
    tidx = threadIdx().x; tidy = threadIdx().y; tidz = threadIdx().z
    bidx = blockIdx().x;  bidy = blockIdx().y;  bidz = blockIdx().z
    dimx = blockDim().x;  dimy = blockDim().y;  dimz = blockDim().z
    c = clock(UInt32)
    
    i = (bidy - 1) * dimy + tidy + ioff
    j = (bidz - 1) * dimz + tidz
    jsx = (j - 1) * size(X, 1)

    if i <= size(I, 1)
        l = ldg(I, i)

        # x = @MVector fill(P[j], maxo ÷ stride + 1)
        x = MVector{maxo ÷ stride + 1, Float32}(undef)
        for o in 0:maxo÷stride
            x[o + 1] = P[l + 1, j, min(size(P, 3), o * stride + tidx)]
        end

        y = @MVector fill(0.0f0, maxo ÷ stride + 1)

        for k = 2:l+1
            for o = 0:maxo÷stride
                ind = ldg(I, (k - 1) * size(I, 1) + i) + o * stride + tidx - 1
                x[o + 1] *= ldg(X, jsx + ind)
                y[o + 1] += ldg(dX, jsx + ind)
            end
        end

        z = warpsum(dot(x, y), stride ÷ 2)
        if tidx == 1
            CUDA.@atomic G[i, j] += z
        end
 
        y = warpsum(sum(x), stride ÷ 2)
        if tidx == 1
            CUDA.@atomic R[i, j] += y
        end
    end

    return
end

function kernel_gradient!(R, G, X, dX, P, I, ioff, roff, ::Val{maxo}, ::Val{stride}) where {maxo, stride}
    tidx = threadIdx().x; tidy = threadIdx().y; tidz = threadIdx().z
    bidx = blockIdx().x;  bidy = blockIdx().y;  bidz = blockIdx().z
    dimx = blockDim().x;  dimy = blockDim().y;  dimz = blockDim().z
    
    i = (bidy - 1) * dimy + tidy + ioff
    j = (bidz - 1) * dimz + tidz + roff

    X = CUDA.Const(X); I = CUDA.Const(I); dX = CUDA.Const(dX); P = CUDA.Const(P)

    @inbounds if i <= size(I, 1)
        l = I[i, 1]

        # c = 0
        x = MVector{maxo ÷ stride + 1, Float32}(undef)
        for o in 0:maxo÷stride
            x[o + 1] = P[l + 1, j, min(size(P, 3), o * stride + tidx)]
        end

        for k = 2:l+1
            # c += 1
            for o = 0:maxo÷stride
                ind = I[i, k] + o * stride + tidx - 1
                x[o + 1] *= X[ind, j]
            end
        end

        for h in 1:size(dX, 3)
            y = @MVector zeros(Float32, maxo ÷ stride + 1)
            for k = 2:l+1
                # c += 1
                for o = 0:maxo÷stride
                    ind = I[i, k] + o * stride + tidx - 1
                    y[o + 1] += dX[ind, j, h]
                end
            end

            z = warpsum(dot(x, y), stride ÷ 2)
            if tidx == 1
                CUDA.@atomic G[i, j - roff, h] += z
            end
        end

        y = warpsum(sum(x), stride ÷ 2)
        # y = warpsum(c, stride ÷ 2)
        if tidx == 1
            CUDA.@atomic R[i, j - roff] += y
        end
    end

    return
end

function applyKernelGradient!(R, G, X::CuArray, dX, P, I, maxoffset, rrange)
    dimx = min(8, maxoffset + 1)
    dimy = 128 ÷ dimx

    for ir in partitionrange(1 : ceil(Int, size(I, 1) / dimy), maxl = 65535)
        blocks = (1, length(ir), length(rrange))
        ioff = (minimum(ir) - 1) * dimy
        roff = minimum(rrange) - 1
        @cuda name="grad" threads=(dimx,dimy,1) blocks=blocks kernel_gradient!(R, G, X, dX, P, I, ioff, roff, Val(maxoffset), Val(dimx))
    end
end


function computelogpGradient!(X, X2, dX, dX2, meanX, XP, dXP, P, R, G, pixelWeights, rotationWeights, images, intensity, npercircle, imagelengths, area; inner = true, mode = :poisson)
    npix = size(X, 1)
    nrot = size(X, 2)
    nimg = size(images, 1)
    lmax = size(images, 2) - 1
    ng = size(G, 3)

    mul!(XP, X', pixelWeights)
    rXP = reshape(XP, 1, size(XP)...)

    P .= .- intensity .* rXP
    maxlogP = maximum(P, dims = (2,3))
    P .= exp.(P .- maxlogP)

    reshape(X2, 2npercircle, :)[1:npercircle, :] .= reshape(X, npercircle, :)
    reshape(X2, 2npercircle, :)[npercircle+1:end, :] .= reshape(X, npercircle, :)
    reshape(dX2, 2npercircle, :)[1:npercircle, :] .= reshape(dX, npercircle, :)
    reshape(dX2, 2npercircle, :)[npercircle+1:end, :] .= reshape(dX, npercircle, :)
    dX2 ./= X2

    for h in 1:size(dX, 3)
        mul!(view(dXP, :, :, h), view(dX, :, :, h)', pixelWeights)
    end

    meanX .= 1
    meanX[2:end, 1:1] .= mean(X2, dims = 2)
    reshaped = reshape(view(meanX, 2:size(meanX, 1), 1), npercircle, :)
    reshaped .= mean(reshaped, dims = 1)
    X2 ./= meanX[2:end]

    result = zeros(nimg,1)
    resultG = zeros(nimg, 1, ng)
    
    for rr in partitionrange(1:nrot, stride = size(R, 2))
        R .= 0; G .= 0
        applyKernelGradient!(R, G, X2, dX2, P, images, ifelse(inner, npercircle-1, 0), rr)

        @views G[:, 1:length(rr), :] .-= intensity .* permutedims(dXP[rr, :, :], (2, 1, 3)) .* R[:, 1:length(rr)]

        view(G, :, 1:length(rr), :) .*= view(rotationWeights, :, rr)
        view(R, :, 1:length(rr)) .*= view(rotationWeights, :, rr)

        result .+= Array(sum(R, dims=2))
        resultG .+= Array(sum(G, dims=2))
    end

    resultG ./= result

    result = log.(result ./ ifelse(inner, npercircle, 1))
    result .+= Array(sum(Float64.(log.(meanX[images[:, 2:end] .+ 1])), dims = 2))
    result .+= getindex.(Ref(reshape(Array(maxlogP), :)), imagelengths .+ 1)

    sum(result), sum(resultG, dims = 1)[:]
end

function gradientComputeClosure(images, rotations, pixels, pixelWeights, npercircle, rotationWeights, N, ndiffs; polarization = 1, area = 1.0, ncalls = 10, rmask = [], nmask = [], usecuda = true, sorted = true, inner = true, mode = :poisson)
    images = if images isa Vector{<:SVector{3, <:Real}}
        deserializeImages(images)
    else
        images
    end

    npix = length(pixels); nrot = length(rotations); 
    nimg = length(images); lmax = maximum(length.(images), init = 0)

    if sorted
        images = sort(images, by = length)
    end
    indexImages = if eltype(images[1]) == Int
        images
    elseif length(eltype(eltype(images))) == 2
        roundImagesCircular(pixels, images, npercircle)
    else
        roundImages(pixels, images)
    end
    indexImages = [image .+ npercircle .* ((image .- 1) .÷ npercircle) for image in indexImages]
    imageMatrix = toImageMatrix(indexImages)
    imagelengths = length.(images)

    I = UInt16.(imageMatrix)
    meanX = ones(2npix + 1, 1)
    X2 = zeros(2npix, nrot)
    dX2 = zeros(2npix, nrot, ndiffs)
    XP = zeros(nrot, size(pixelWeights, 2))
    dXP = zeros(nrot, size(pixelWeights, 2), ndiffs)
    P = zeros(lmax+1, nrot, size(pixelWeights, 2))
    R = zeros(nimg, ceil(Int, nrot / ncalls))
    G = zeros(nimg, ceil(Int, nrot / ncalls), ndiffs)
    pixelWeights = pixelWeights
    rotationWeights = reshape(copy(rotationWeights),1,:)

    cuI = cu(I); cumeanX = cu(meanX); cuX2 = cu(X2); cuXP = cu(XP); cuP = cu(P); cuR = cu(R);
    cupixelWeights = cu(pixelWeights); curotationWeights = cu(rotationWeights); cuG = cu(G)
    cudX2 = cu(dX2); cudXP = cu(dXP); 

    function setimages!(images)
        if maximum(length, images) > lmax
            throw(BoundsError("To many photons"))
        end
        if sorted
            images = sort(images, by = length)
        end
        indexImages = if eltype(images[1]) == Int
            images
        elseif length(eltype(eltype(images))) == 2
            roundImagesCircular(pixels, images, npercircle)
        else
            roundImages(pixels, images)
        end
        indexImages = [image .+ npercircle .* ((image .- 1) .÷ npercircle) for image in indexImages]

        imageMatrix = toImageMatrix(indexImages, lmax)

        copyto!(cuI, UInt16.(imageMatrix))
    end


    grad(X, dX) = computelogpGradient!(X, cuX2, dX, cudX2, cumeanX, cuXP, cudXP, cuP, cuR, cuG, cupixelWeights, curotationWeights, cuI, N, npercircle, imagelengths, area, inner = inner, mode = mode)
    grad, setimages!
end


function gradientClosure(images, rotations, pixels, pixelWeights, npercircle, rotationWeights, N, ndiffs; polarization = 1, area = 1.0, ncalls = 10, rmask = [], nmask = [], usecuda = true, sorted = true, inner = true, mode = :poisson)
    eval = setupEval(rotations, pixels, usecuda = usecuda)
    evalGrad = setupGradientEval(rotations, pixels, ndiffs, usecuda = usecuda)
    
    computegradient!, setimages! = gradientComputeClosure(images, rotations, pixels, pixelWeights, npercircle, rotationWeights, N, ndiffs; ncalls = ncalls, sorted = sorted)

    function grad(f)
        X = eval(f)
        dX = evalGrad(f)
        lp, lg = computegradient!(X, dX)
        g = SVector{3, Float64}.(eachcol(reshape(lg, 3, :)))
        lp, g
    end
    
    grad, setimages!
end

function setupGradientEval(rotations, pixels, ndiffs; usecuda = true)
    dX = CUDA.zeros(length(pixels), length(rotations), ndiffs)
    points = cu(SVector{3, Float32}.([R * x for x in pixels, R in rotations]))

    function ev(f)
        for h in 1:length(f)
            for i in 1:3
                ind = 3 * (h - 1) + i
                ind > ndiffs && break
                dX[:, :, ind] .= getindex.(gradient.(cu(f), h, points), i)
            end
        end
        dX
    end
end

function gradientClosure(images, params, N, ndiffs; kwargs...)
    gradientClosure(images, params.rotations, params.pixels, params.pixelWeights, params.npercircle, params.rotationWeights, N, ndiffs; kwargs...)
end

function gradientClosure(images, N, ndiffs; kwargs...)
    args, kwargs = convertArguments(images, N; kwargs...)
    gradientClosure(args..., ndiffs; kwargs...)
end


