abstract type Proposer end

struct NormalProposer <: Proposer

end

function neighbor(proposer::NormalProposer, f, d)
    f .+ d .* randn.(eltype(f))
end

function logprop(proposer::NormalProposer, f_to, f_from, d)\
    mapreduce(+, f_to, f_from) do x, y
        norm(x - y)^2 / (2d^2)
    end
end


struct SmoothProposer <: Proposer
    σ::Float64

    function SmoothProposer(σ = 1.0)
        new(σ)
    end
end

function neighbor(proposer::SmoothProposer, f, d)
    V = exp.(-1/(2*proposer.σ^2) .* pairwise(Euclidean(), f.positions).^2)
    V = 0.9V + 0.1I
    D = MvNormal(zeros(3length(f)), kron(V, Matrix(I, 3, 3)))
    s = d .* eltype(f.positions).(eachcol(reshape(rand(D), 3, :)))
    s .-= Ref(mean(s))

    newf = deepcopy(f)
    newf.positions .+= s
    newf
end

function logprop(proposer::SmoothProposer, f_to, f_from, d)
    V = exp.(-1/(2*proposer.σ^2) .* pairwise(Euclidean(), f_from.positions).^2)
    V = 0.9V + 0.1I
    D = MvNormal(zeros(3length(f_from)), kron(V, Matrix(I, 3, 3)))
    s = (f_to.positions .- f_from.positions) ./ d
    logpdf(D, reduce(vcat, s))
end


struct StructureProposer{T} <: Proposer
    gnormals::T

    function StructureProposer(g, σ = 0.0)
        gnormals = [MvNormal(g.positions[i], sqrt(g.widths[i]^2 + σ^2)) for i in 1:length(g)]
        new{typeof(gnormals)}(gnormals)
    end
end

function StructureProposer(w::Real, σ = 0.0)
    StructureProposer(AtomVolume(1, w))
end

function neighbor(proposer::StructureProposer, f, d)
    fnormals = [proposer.gnormals[ceil(Int, length(proposer.gnormals) / length(f) * i)] for i in 1:length(f)]
    normals = [normalprod(fnormals[i], MvNormal(f.positions[i], d)) for i in 1:length(f)]

    newf = deepcopy(f)
    newf.positions .= [rand(n) for n in normals]
    newf.positions .-= Ref(mean(newf.positions))
    
    newf
end

function logprop(proposer::StructureProposer, f_to, f_from, d)
    fnormals = [proposer.gnormals[ceil(Int, length(proposer.gnormals) / length(f_from) * i)] for i in 1:length(f_from)]
    normals = [normalprod(fnormals[i], MvNormal(f_from.positions[i], d)) for i in 1:length(f_from)]

    sum(log.(pdf.(normals, f_to.positions)))
end

struct CombinedProposer{T} <: Proposer
    proposers::T
end

CombinedProposer(proposers...) = CombinedProposer(proposers)

function neighbor(proposer::CombinedProposer, x, d)
    y = x
    for p in proposer.proposers
        y = neighbor(p, y, d)
    end
    y
end

function logprop(proposer::CombinedProposer, x, y, d)
    sum(logprop(p, x, y, d) for p in proposer.proposers)
end


struct MultiProposer{T} <: Proposer
    proposers::T
end

function neighbor(proposer::MultiProposer, fs, d)
    [neighbor(p, f, d) for (p, f) in zip(proposer.proposers, fs)]
end

function logprop(proposer::MultiProposer, fs, gs, d)
    sum(logprop(p, f, g, d) for (p, f, g) in zip(proposer.proposers, fs, gs))
end


struct GibbsProposer{T} <: Proposer
    proposer::T
    i::Int
end

function neighbor(proposer::GibbsProposer, f, d)
    @set f[proposer.i] = neighbor(proposer.proposer, f[proposer.i], d)
end

function logprop(proposer::GibbsProposer, f, g, d)
    logprop(proposer.proposer, f[proposer.i], g[proposer.i], d)
end


struct WeightProposer{T} <: Proposer
    prior::T
end

# function neighbor(proposer::WeightProposer, ws, d)
#     # d = min(d, 2std(proposer.prior))
#     wprior(ws) = prod(pdf.(proposer.prior, ws))
#     for i in 1:100000
#         news = ws .+ d .* (x = randn(length(ws)); x .- mean(x))
#         if all(0 .< news .< 1) && rand() < wprior(news) / wprior(ws)
#             ws = news
#             break
#         end
#     end
#     normalize!(ws, 1)
#     ws = clamp.(ws, 0, 1)
#     ws
# end

function metropolisgamma(x, d)
    while true
        y = x * exp(randn() * d)
        if rand() < exp(-y) / exp(-x) * y / x
            return y
        end
    end
end

function neighbor(proposer::WeightProposer, ws, d)
    wprior(ws) = prod(pdf.(proposer.prior, ws ./ sum(ws) .* length(ws)))

    news = metropolisgamma.(ws, d)
    # news = [rand(Gamma(1 / d^2, w * d^2)) for w in ws]
    if rand() < wprior(news) / wprior(ws)
        ws = news
    end
    ws
end

function logprop(proposer::WeightProposer, fws, gws, d)
    wprior(ws) = prod(pdf.(proposer.prior, ws ./ sum(ws) .* length(ws)))
    log(wprior(fws)) - log(wprior(gws))
end

struct WidthProposer <: Proposer
    d::Float64
end

function neighbor(proposer::WidthProposer, fs::Array, d)
    newfs = deepcopy(fs)
    rd = randn() .* proposer.d
    for newf in newfs
        newf.widths .+= rd
        newf.widths .= clamp.(newf.widths, 1e-2, Inf)
    end
    newfs
end

function neighbor(proposer::WidthProposer, f, d)
    newf = deepcopy(f)
    rd = randn() .* proposer.d
    newf.widths .+= rd
    newf.widths .= clamp.(newf.widths, 1e-2, Inf)
    newf
end


function logprop(proposer::WidthProposer, f, g, d)
    0.0
end

struct HeightProposer <: Proposer
    minh::Float64
    maxh::Float64
end

function neighbor(proposer::HeightProposer, f, d)
    newf = deepcopy(f)
    newf.heights .+= randn.() .* d
    newf.heights .= clamp.(newf.heights, proposer.minh, proposer.maxh)
    newf
end


function logprop(proposer::HeightProposer, f, g, d)
    0.0
end

struct SmartProposer <: Proposer
    structureproposer::Union{Nothing, StructureProposer}
    weightproposer::Union{Nothing, WeightProposer}
    widthproposer::Union{Nothing, WidthProposer}
    heightproposer::Union{Nothing, HeightProposer}
end

function SmartProposer(; σ = nothing, widthstep = nothing, weightprior = nothing, minh = nothing, maxh = nothing)
    structureproposer = if isnothing(σ)
        nothing
    elseif σ isa Real
        StructureProposer(AtomVolume(1, σ))
    elseif σ isa AtomVolume
        StructureProposer(σ)
    else
        σ
    end
    weightproposer = isnothing(weightprior) ? nothing : WeightProposer(weightprior)
    widthproposer = isnothing(widthstep) ? nothing : WidthProposer(widthstep)
    heightproposer = isnothing(minh) ? nothing : HeightProposer(minh, maxh)
    SmartProposer(structureproposer, weightproposer, widthproposer, heightproposer)
end

function neighbor(p::SmartProposer, x::AtomVolume, d)
    y = neighbor(p.structureproposer, x, d)
    y = neighbor(p.widthproposer, y, d)
    y = neighbor(p.heightproposer, y, d)
    y
end
neighbor(p::SmartProposer, x::MixtureVolume, d) = (@set x.fs[1] = neighbor(p, x.fs[1], d))
neighbor(p::SmartProposer, x::Vector{<:AbstractFloat}, d) = neighbor(p.weightproposer, x, d)
neighbor(p::SmartProposer, x, d) = neighbor.(Ref(p), x, d)

function logprop(p::SmartProposer, x::AtomVolume, y::AtomVolume, d)
    logprop(p.widthproposer, x, y, d) + logprop(p.structureproposer, x, y, d) + logprop(p.heightproposer, x, y, d)
end
logprop(p::SmartProposer, x::MixtureVolume, y::MixtureVolume, d) = logprop(p, x.fs[1], y.fs[1], d)
logprop(p::SmartProposer, x::Vector{<:AbstractFloat}, y::Vector{<:AbstractFloat}, d) = logprop(p.weightproposer, x, y, d)
logprop(p::SmartProposer, x, y, d) = sum(logprop.(Ref(p), x, y, d))

neighbor(p::Nothing, x, d) = x
logprop(p::Nothing, x, y, d) = 0.0

const WiP = WidthProposer
const WeP = WeightProposer
const GP = GibbsProposer
const MP = MultiProposer
const SP = StructureProposer
const NP = NormalProposer
const CP = CombinedProposer
